import os
import sys
import time
import timeit
from pathlib import Path, PurePosixPath

import h5py
import numpy as np
from tqdm import tqdm

from .h5 import h5_print_contents
from .utils import even_print, format_time_string

# ======================================================================

def _calc_statistics_wall(self, **kwargs):
    '''
    Calculate statistics for an unsteady wall measurement
      which has dimensions (nx,1,nz)
    - mean
    - covariance
    - skewness, kurtosis
    - probability distribution function (PDF)
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'spd.calc_statistics_wall()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if not self.usingmpi:
        raise NotImplementedError('function is not implemented for non-MPI usage')
    
    h5py_is_mpi_build = h5py.h5.get_config().mpi
    if not h5py_is_mpi_build:
        if verbose: print('h5py was not compiled for parallel usage! exiting.')
        sys.exit(1)
    
    ri = kwargs.get('ri',1)
    rj = kwargs.get('rj',1)
    rt = kwargs.get('rt',1)
    
    fn_h5_stats = kwargs.get('fn_h5_stats',None) ## filename for output HDF5 (.h5) file
    n_bins      = kwargs.get('n_bins',512) ## n bins for histogram (PDF) calculation
    
    ## check data distribution
    if (rj!=1):
        raise AssertionError('rj!=1')
    if (rt!=1):
        raise AssertionError('rt!=1')
    if (ri*rj*rt != self.n_ranks):
        raise AssertionError('ri*rj*rt != self.n_ranks')
    if (ri>self.ni):
        raise AssertionError('ri>self.ni')
    if (self.ni%ri!=0):
        raise AssertionError('ni currently needs to be divisible by the n ranks')
    
    ## distribute data over i/[x]
    ril_ = np.array_split(np.arange(self.ni,dtype=np.int64),min(ri,self.ni))
    ril = [[b[0],b[-1]+1] for b in ril_ ]
    ri1,ri2 = ril[self.rank]
    nir = ri2 - ri1
    
    ## output filename : HDF5 (.h5)
    if (fn_h5_stats is None): ## automatically determine name
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        #fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fname_stats_h5_base = fname_root+'_stats.h5'
        fn_h5_stats = str(PurePosixPath(fname_path, fname_stats_h5_base))
    if (Path(fn_h5_stats).suffix != '.h5'):
        raise ValueError(f"fn_h5_stats='{str(fn_h5_stats)}' must end in .h5")
    if os.path.isfile(fn_h5_stats):
        if (fn_h5_stats == self.fname):
            raise ValueError(f"fn_h5_stats='{str(fn_h5_stats)}' cannot be same as input filename.")
    
    if verbose: even_print( 'fn_h5'        , self.fname   )
    if verbose: even_print( 'fn_h5_stats'  , fn_h5_stats  )
    if verbose: print(72*'-')
    if verbose: even_print( 'ni' , f'{self.ni:d}' )
    if verbose: even_print( 'nj' , f'{self.nj:d}' )
    if verbose: even_print( 'nt' , f'{self.nt:d}' )
    if verbose: print(72*'-')
    self.comm.Barrier()
    
    # ===
    
    ## the data dictionary to be pickled or written to .h5 later
    data = {}
    
    ## freestream data
    data['lchar']   = self.lchar
    data['U_inf']   = self.U_inf
    data['rho_inf'] = self.rho_inf
    data['T_inf']   = self.T_inf
    data['mu_inf']  = self.mu_inf
    data['p_inf']   = self.p_inf
    data['Ma']      = self.Ma
    data['Pr']      = self.Pr
    
    ## read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    data['x'] = x
    data['z'] = z
    data['t'] = t
    
    nx = self.ni
    ni = self.ni
    nz = self.nj
    nj = self.nj
    data['ni'] = ni
    data['nx'] = nx
    data['nj'] = nj
    data['nz'] = nz
    
    nt = self.nt
    data['nt'] = nt
    
    ## assert constant Δz
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-6)):
        raise NotImplementedError('Δz not constant')
    
    ## get Δt, dimensional [s]
    if hasattr(self,'dt'):
        dt = self.dt * self.tchar
        np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-12, atol=1e-12)
    else:
        dt = t[1] - t[0] ## already dimensional
    
    if hasattr(self,'duration'):
        t_meas = self.duration * self.tchar
        np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-12, atol=1e-12)
    else:
        t_meas = t[-1] - t[0] ## already dimensional
        self.duration = t_meas / self.tchar ## non-dimensionalize for attribute
    
    zrange = z.max() - z.min()
    
    data['t'] = t
    data['dt'] = dt
    data['dz'] = dz0
    data['zrange'] = zrange
    
    if verbose: even_print( 'Δt/tchar'       , f'{dt/self.tchar:0.8f}' )
    if verbose: even_print( 'Δt'             , f'{dt:0.3e} [s]'        )
    if verbose: even_print( 'duration/tchar' , f'{self.duration:0.1f}' )
    if verbose: even_print( 'duration'       , f'{self.duration*self.tchar:0.3e} [s]' )
    if verbose: print(72*'-')
    
    ## report
    if verbose:
        even_print('Δt'     , f'{dt    :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas:0.5e} [s]' )
        even_print('Δz'     , f'{dz0   :0.5e} [m]' )
        even_print('zrange' , f'{zrange:0.5e} [m]' )
        #print(72*'-')
    
    # ==============================================================
    # prepare buffers, maps, etc.
    # ==============================================================
    
    ## key    = str:scalar name
    ## value  = tuple:recipe,
    ##          bool:do_mean,
    ##          bool:do_pdf,
    ##          bool:do_skew_kurt
    ##
    ## recipe elements:
    ##    - if tuple : ( str:scalar , bool:ρ weighting ) --> always mean-removed
    ##    - if str   : str:scalar
    
    scalars_dict = {
        
        'u_tau'     : [ ( 'u_tau',                  ) ,  True, True, False ], ## uτ
        'utauI'     : [ (          ('u_tau',False), ) , False, True,  True ], ## uτ′
        'r_utauII'  : [ (   'rho', ('u_tau',True ), ) , False, True,  True ], ## ρ·uτ″
        
        'v_tau'     : [ ( 'v_tau',                  ) ,  True, True, False ], ## vτ
        'vtauI'     : [ (          ('v_tau',False), ) , False, True,  True ], ## vτ′
        'r_vtauII'  : [ (   'rho', ('v_tau',True ), ) , False, True,  True ], ## ρ·vτ″
        
        'w_tau'     : [ ( 'w_tau',                  ) ,  True, True, False ], ## wτ
        'wtauI'     : [ (          ('w_tau',False), ) , False, True,  True ], ## wτ′
        'r_wtauII'  : [ (   'rho', ('w_tau',True ), ) , False, True,  True ], ## ρ·wτ″
        
        'utauIutauI' : [ ( ('u_tau',False), ('u_tau',False) ), True, True, False ], ## uτ′uτ′
        'vtauIvtauI' : [ ( ('v_tau',False), ('v_tau',False) ), True, True, False ], ## vτ′vτ′
        'wtauIwtauI' : [ ( ('w_tau',False), ('w_tau',False) ), True, True, False ], ## wτ′wτ′
        #'utauIvtauI' : [ ( ('u_tau',False), ('v_tau',False) ), True, True, False ], ## uτ′vτ′
        #'utauIwtauI' : [ ( ('u_tau',False), ('w_tau',False) ), True, True, False ], ## uτ′wτ′
        
        'tau_uy'       : [ (  'tau_uy',                          ) ,  True, True, False ], ## τuy
        'tauuyI'       : [ ( ('tau_uy',False) ,                  ) , False, True,  True ], ## τ′uy
        'tauuyItauuyI' : [ ( ('tau_uy',False) , ('tau_uy',False) ) ,  True, True, False ], ## τ′uy·τ′uy
        
        'tau_vy'       : [ (  'tau_vy',                          ) ,  True, True, False ], ## τvy
        'tauvyI'       : [ ( ('tau_vy',False) ,                  ) , False, True,  True ], ## τ′vy
        'tauvyItauvyI' : [ ( ('tau_vy',False) , ('tau_vy',False) ) ,  True, True, False ], ## τ′vy·τ′vy
        
        'tau_wy'       : [ (  'tau_wy',                          ) ,  True, True, False ], ## τwy
        'tauwyI'       : [ ( ('tau_wy',False) ,                  ) , False, True,  True ], ## τ′wy
        'tauwyItauwyI' : [ ( ('tau_wy',False) , ('tau_wy',False) ) ,  True, True, False ], ## τ′wy·τ′wy
        
        'TITI'     : [ (        ('T',False), ('T',False) ) , True, True, False ], ## T′T′
        'TIITII'   : [ (        ('T',True ), ('T',True ) ) , True, True, False ], ## T″T″
        'r_TIITII' : [ ( 'rho', ('T',True ), ('T',True ) ) , True, True, False ], ## ρ·T″T″
        
        'rho'      : [ (  'rho',                       ) ,  True, True, False, ], ## ρ
        'rhoI'     : [ (                ('rho',False), ) , False, True,  True, ], ## ρ′
        'rhoIrhoI' : [ ( ('rho',False), ('rho',False), ) ,  True, True, False, ], ## ρ′ρ′
        
        'p'    : [ (  'p',                     ) ,  True, True, False, ], ## p
        'pI'   : [ (              ('p',False), ) , False, True,  True, ], ## p′
        'pIpI' : [ ( ('p',False), ('p',False), ) ,  True, True, False, ], ## p′p′
        
        'T'        : [ ( 'T',                ) ,  True, True, False, ], ## T
        'TI'       : [ (        ('T',False), ) , False, True,  True, ], ## T′
        'TII'      : [ (        ('T',True ), ) , False, True, False, ], ## T″
        'r_TII'    : [ ( 'rho', ('T',True ), ) , False, True,  True, ], ## ρ·T″
        
        }
    
    scalars_avg  = []
    scalars_pdf  = []
    scalars_hos_ = []
    for s,ss in scalars_dict.items():
        recipe, do_mean, do_pdf, do_skew_kurt = ss
        if do_mean:
            scalars_avg.append(s)
        if do_pdf:
            scalars_pdf.append(s)
        if do_skew_kurt:
            scalars_hos_.append(s)
    
    scalars_hos=[]
    for s_ in ['skew','kurt']:
        for ss_ in scalars_hos_:
            scalars_hos.append(f'{ss_}_{s_}')
    
    ## rank-local buffsres
    data_avg  = np.zeros(shape=(nir,),          dtype={'names':scalars_avg, 'formats':[ np.float64 for sss in scalars_avg ]})
    data_bins = np.zeros(shape=(nir,n_bins+1),  dtype={'names':scalars_pdf, 'formats':[ np.float64 for sss in scalars_pdf ]})
    data_pdf  = np.zeros(shape=(nir,n_bins),    dtype={'names':scalars_pdf, 'formats':[ np.float64 for sss in scalars_pdf ]})
    data_hos  = np.zeros(shape=(nir,),          dtype={'names':scalars_hos, 'formats':[ np.float64 for sss in scalars_hos ]})
    
    # ==============================================================
    # main loop
    # ==============================================================
    
    if verbose:
        progress_bar = tqdm(
            total=len(scalars_dict)*nir,
            ncols=100,
            desc='statistics',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#FF6600',
            )
    
    for s,ss in scalars_dict.items(): ## e.g. where s='r_uIIvII' & ss=[ ( 'rho', ('u',True ), ('v',True ) ), True,True,False ]
        
        if verbose: tqdm.write(72*'-')
        
        recipe, do_mean, do_pdf, do_skew_kurt = ss
        
        if verbose:
            tqdm.write(even_print('computing',s,s=True,))
        
        ## should ρ be read?
        read_rho = False
        for s_ in recipe:
            if isinstance(s_, str) and (s_=='rho'):
                read_rho = True
            elif isinstance(s_, str) and (s_!='rho'):
                pass
            elif isinstance(s_, tuple):
                if not isinstance(s_[1], bool):
                    raise ValueError
                if s_[1]: ## i.e. density-weighted
                    read_rho = True
            else:
                raise ValueError
        
        ## [x] loop (rank-local)
        for ii in range(ri1,ri2):
            
            iii = ii - ri1
            
            ## read ρ
            if read_rho:
                
                dset = self['data/rho']
                self.comm.Barrier()
                t_start = timeit.default_timer()
                if self.usingmpi:
                    with dset.collective:
                        rho = np.copy( dset[ii,:,:] )
                else:
                    rho = np.copy( dset[ii,:,:] )
                self.comm.Barrier()
                t_delta = timeit.default_timer() - t_start
                data_gb = self.n_ranks * 1 * self.nj * self.nt * dset.dtype.itemsize / 1024**3
                if verbose:
                    tqdm.write(even_print('read: ρ', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
                
                rho = rho.astype(np.float64) ## cast to double
                
                ## re-dimensionalize
                rho *= self.rho_inf
                
                if ( rho.shape != (nj,nt) ):
                    raise ValueError
                    print(f'rank {self.rank:d}: shape violation')
                    self.comm.Abort(1)
                
                ## ρ mean in [t] --> leave (j,1)
                rho_avg = np.mean(rho, axis=-1, dtype=np.float64, keepdims=True)
            
            else:
                rho     = None ; del rho
                rho_avg = None ; del rho_avg
            
            ## product buffer for multiplication --> !!! notices ones() here and not zeros() !!!
            #data_accum = np.ones(shape=(ni,nj,nt), dtype=np.float64) ## chunk range context
            data_accum = np.ones(shape=(nj,nt), dtype=np.float64) ## chunk range context
            
            ## read unsteady scalar data, remove mean, <density weight>, multiply
            for sss in recipe:
                
                if isinstance(sss, str) and (sss=='rho'): ## ρ
                    
                    ## multiply product accumulator, ρ was already read and is already dimensional
                    data_accum *= rho
                
                elif isinstance(sss, str) and (sss!='rho'): ## scalar which will NOT be mean-removed
                    
                    ## read
                    dset = self[f'data/{sss}']
                    self.comm.Barrier()
                    t_start = timeit.default_timer()
                    if self.usingmpi:
                        with dset.collective:
                            data_X = np.copy( dset[ii,:,:] )
                    else:
                        data_X = np.copy( dset[ii,:,:] )
                    self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    data_gb = self.n_ranks * 1 * self.nj * self.nt * dset.dtype.itemsize / 1024**3
                    if verbose:
                        tqdm.write(even_print(f'read: {sss}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
                    
                    data_X = data_X.astype(np.float64) ## cast to double
                    
                    if ( data_X.shape != (nj,nt) ):
                        raise ValueError
                        print(f'rank {self.rank:d}: shape violation')
                        self.comm.Abort(1)
                    
                    ## redimensionalize
                    if sss in ['tau_uy','tau_vy','tau_wy']:
                        data_X *= self.rho_inf * self.U_inf**2
                    elif sss in ['u_tau','v_tau','w_tau']:
                        data_X *= self.U_inf
                    elif sss in ['T',]:
                        data_X *= self.T_inf
                    elif sss in ['rho',]:
                        data_X *= self.rho_inf
                    elif sss in ['p',]:
                        data_X *= self.rho_inf * self.U_inf**2
                    else:
                        raise ValueError(f"condition needed for redimensionalizing '{str(sss)}'")
                    
                    ## MULTIPLY product accumulator
                    data_accum *= data_X
                
                elif isinstance(sss, tuple): ## scalar which WILL be mean-removed (with- or without ρ-weighting)
                    
                    if (len(sss)!=2):
                        raise ValueError
                    
                    sn, do_density_weighting = sss ## e.g. ('u',True)
                    
                    ## read
                    dset = self[f'data/{sn}']
                    self.comm.Barrier()
                    t_start = timeit.default_timer()
                    if self.usingmpi:
                        with dset.collective:
                            data_X = np.copy( dset[ii,:,:] )
                    else:
                        data_X = np.copy( dset[ii,:,:] )
                    self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    data_gb = self.n_ranks * 1 * self.nj * self.nt * dset.dtype.itemsize / 1024**3
                    if verbose:
                        tqdm.write(even_print(f'read: {sn}', '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
                    
                    data_X = data_X.astype(np.float64) ## cast to double
                    
                    ## redimensionalize
                    if sn in ['tau_uy','tau_vy','tau_wy']:
                        data_X *= self.rho_inf * self.U_inf**2
                    elif sn in ['u_tau','v_tau','w_tau']:
                        data_X *= self.U_inf
                    elif sn in ['T',]:
                        data_X *= self.T_inf
                    elif sn in ['rho',]:
                        data_X *= self.rho_inf
                    elif sn in ['p',]:
                        data_X *= self.rho_inf * self.U_inf**2
                    else:
                        raise ValueError(f"condition needed for redimensionalizing '{str(sn)}'")
                    
                    ## avg(□) or avg(ρ·□)/avg(ρ)
                    if do_density_weighting:
                        data_X_mean = np.mean( rho*data_X , axis=-1, dtype=np.float64, keepdims=True) ## (nz,1)
                        data_X_mean /= rho_avg
                    else:
                        data_X_mean = np.mean( data_X , axis=-1, dtype=np.float64, keepdims=True) ## (nz,1)
                    
                    ## Reynolds prime □′ or Favre prime □″
                    data_X -= data_X_mean
                    
                    ## assert avg(□′)==0 or avg(ρ·□″)==0
                    if do_density_weighting:
                        a_ = np.mean(rho*data_X, axis=-1, dtype=np.float64, keepdims=True)
                    else:
                        a_ = np.mean(data_X, axis=-1, dtype=np.float64, keepdims=True)
                    if not np.allclose( a_, np.zeros_like(a_), atol=1e-3 ):
                        print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                        self.comm.Abort(1)
                    
                    ## MULTIPLY product accumulator by □′ | □″
                    data_accum *= data_X
                
                else:
                    raise ValueError
            
            # ===============================================================================
            # At this point you have the UNSTEADY 2D [j,t] data according to 'recipe'
            # ===============================================================================
            
            #xiA = cx1 - rx1
            #xiB = cx2 - rx1
            
            self.comm.Barrier()
            t_start = timeit.default_timer()
            
            ## mean in [t] --> leave [j,1]
            d_mean = np.mean( data_accum, axis=-1, keepdims=True, dtype=np.float64)
            #if ( d_mean.shape != (ni,nj,1,) ):
            if ( d_mean.shape != (nj,1) ):
                raise ValueError
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            if do_mean: ## mean in [j] --> leave float
                data_avg[s][iii] = np.mean( d_mean, axis=(0,1), dtype=np.float64)
            
            if do_pdf:
                
                d_ = np.copy( data_accum.ravel() )
                if ( d_.shape != (nj*nt,) ):
                    raise ValueError
                    print(f'rank {self.rank:d}: shape violation')
                    self.comm.Abort(1)
                pdf_ , bin_edges_ = np.histogram( d_ , bins=n_bins , density=True )
                data_bins[s][iii,:] = bin_edges_
                data_pdf[s][iii,:]  = pdf_
            
            if do_skew_kurt:
                
                d_ = np.copy( data_accum.ravel() ) ## □′ or □″, although mean ==0 is not checked
                if ( d_.shape != (nj*nt,) ):
                    raise ValueError
                    print(f'rank {self.rank:d}: shape violation')
                    self.comm.Abort(1)
                
                ## ## is it interpretable as a Reynolds mean-removed quantity? (□′)
                ## interpretable_as_Re_mean = False
                ## a_ = np.mean(d_, axis=3, dtype=np.float64, keepdims=True) ## mean in [t]
                ## np.testing.assert_equal( a_.shape, (nx,1,nz,1,) )
                ## if np.allclose(a_, np.zeros_like(a_), atol=1e-10):
                ##     interpretable_as_Re_mean = True
                ## 
                ## ## is it interpretable as a Favre mean-removed quantity? (□″)
                ## interpretable_as_Fv_mean = False
                ## if read_rho and not interpretable_as_Re_mean:
                ##     #rho_ = np.copy( rho[:,yii,:,:] )
                ##     rho_ = np.zeros(shape=(nx,1,nz,nt), dtype=dtype_unsteady)
                ##     rho_[:,:,:,:] = rho[:,yii,:,:][:,np.newaxis,:,:] ## [x,1,z,t]
                ##     np.testing.assert_equal( rho_.shape, (nx,1,nz,nt) )
                ##     
                ##     a_ = np.mean(d_*rho_, axis=3, dtype=np.float64, keepdims=True) ## mean in [t]
                ##     np.testing.assert_equal( a_.shape, (nx,1,nz,1,) )
                ##     if np.allclose(a_, np.zeros_like(a_), atol=1e-10):
                ##         interpretable_as_Fv_mean = True
                ## 
                ## if (not interpretable_as_Re_mean) and (not interpretable_as_Fv_mean):
                ##     print(f'{s} has non-zero Favre and Reynolds mean, i.e. is not <□′>!=0 and <ρ□″>!=0')
                ##     self.comm.Abort(1)
                
                d_std = np.sqrt( np.mean( d_**2 , dtype=np.float64 ) )
                
                if np.isclose(d_std, 0., atol=1e-08):
                    d_skew = 0.
                    d_kurt = 0.
                else:
                    d_skew = np.mean( d_**3 , dtype=np.float64 ) / d_std**3
                    d_kurt = np.mean( d_**4 , dtype=np.float64 ) / d_std**4
                
                data_hos[f'{s}_skew'][iii] = d_skew
                data_hos[f'{s}_kurt'][iii] = d_kurt
            
            self.comm.Barrier() ## [x] loop ('ii' within this rank's range)
            if verbose: progress_bar.update()
    
    self.comm.Barrier()
    if verbose:
        progress_bar.close()
        print(72*'-')
    
    # ==============================================================
    # write HDF5 (.h5) file
    # ==============================================================
    
    ## open on rank 0 and write attributes, dimensions, etc.
    if (self.rank==0):
        with h5py.File(fn_h5_stats, 'w') as hfw:
            
            ## write floats,ints as top-level attributes
            for key,val in data.items():
                if isinstance(data[key], (int,np.int32,np.int64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], (float,np.float32,np.float64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], np.ndarray):
                    pass
                else:
                    print(f'key {key} is type {str(type(data[key]))}')
                    self.comm.Abort(1)
            
            ## write numpy arrays
            hfw.create_dataset( 'dims/x' , data=x ) ## [m]
            hfw.create_dataset( 'dims/z' , data=z ) ## [m]
            hfw.create_dataset( 'dims/t' , data=t ) ## [s]
            
            ## initialize datasets
            for scalar in scalars_avg:
                hfw.create_dataset( f'avg/{scalar}', shape=(nx,), dtype=np.float64, chunks=(1,) )
            for scalar in scalars_pdf:
                hfw.create_dataset( f'bins/{scalar}', shape=(nx,n_bins+1), dtype=np.float64, chunks=(1,n_bins+1) )
            for scalar in scalars_pdf:
                hfw.create_dataset( f'pdf/{scalar}', shape=(nx,n_bins), dtype=np.float64, chunks=(1,n_bins) )
            for scalar in scalars_hos:
                hfw.create_dataset( f'hos/{scalar}', shape=(nx,), dtype=np.float64, chunks=(1,) )
    
    self.comm.Barrier()
    time.sleep(2.)
    self.comm.Barrier()
    
    with h5py.File(fn_h5_stats, 'a', driver='mpio', comm=self.comm) as hfw:
        
        # === collectively write data
        
        for scalar in scalars_avg:
            dset = hfw[f'avg/{scalar}']
            with dset.collective:
                dset[ri1:ri2] = data_avg[scalar][:]
            self.comm.Barrier()
        
        for scalar in scalars_pdf:
            dset = hfw[f'bins/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = data_bins[scalar][:,:]
            self.comm.Barrier()
        
        for scalar in scalars_pdf:
            dset = hfw[f'pdf/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = data_pdf[scalar][:,:]
            self.comm.Barrier()
        
        for scalar in scalars_hos:
            dset = hfw[f'hos/{scalar}']
            with dset.collective:
                dset[ri1:ri2] = data_hos[scalar][:]
            self.comm.Barrier()
    
    ## report file size
    if verbose:
        even_print( os.path.basename(fn_h5_stats), f'{(os.path.getsize(fn_h5_stats)/1024**2):0.2f} [MB]' )
        print(72*'-')
    
    ## report file contents
    self.comm.Barrier()
    if (self.rank==0):
        with h5py.File(fn_h5_stats,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : spd.calc_statistics_wall() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return
    


import os
import timeit

import numpy as np
import scipy as sp

from .bl import calc_bl_integral_quantities_1d, calc_d99_1d, calc_profile_edge_1d
from .gradient import gradient
from .utils import even_print, format_time_string

# ======================================================================

def _add_mean_dimensional_data_xpln(self, **kwargs):
    '''
    Get dimensionalized mean data for [x] plane
    --> save to existing RGD file with fsubtype=mean
    - assumes volume which is thin in [x] direction
    - an RGD which is the output of rgd.get_mean() should be opened here
    - NOT parallel!!
    '''
    
    verbose      = kwargs.get('verbose',True)
    epsilon      = kwargs.get('epsilon',5e-5)
    acc          = kwargs.get('acc',6)
    edge_stencil = kwargs.get('edge_stencil','full')
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'rgd.add_mean_dimensional_data_xpln()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    ## This func is not parallel!
    if self.usingmpi:
        raise NotImplementedError('rgd.add_mean_dimensional_data_xpln() is not a parallel function')
    
    ## 'r' and 'w' open modes are not allowed
    if not (self.open_mode=='a') or (self.open_mode=='r+'):
        raise ValueError(f'open mode is {self.open_mode}')
    
    ## Assert that this is a mean flow file ( i.e. output from rgd.get_mean() )
    if (self.fsubtype!='mean'):
        print(self.fsubtype)
        raise ValueError
    
    ## Get size of infile
    fsize = os.path.getsize(self.fname)/1024**3
    if verbose: even_print(os.path.basename(self.fname),'%0.1f [GB]'%fsize)
    if verbose: even_print('nx','%i'%self.nx)
    if verbose: even_print('ny','%i'%self.ny)
    if verbose: even_print('nz','%i'%self.nz)
    if verbose: even_print('nt','%i'%self.nt)
    if verbose: even_print('ngp','%0.1f [M]'%(self.ngp/1e6,))
    if verbose: print(72*'-')
    
    ## Read in 1D coordinate arrays, then dimensionalize [m]
    x = np.copy( self['dims/x'][()] * self.lchar )
    y = np.copy( self['dims/y'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    
    if (x.ndim!=1):
        raise ValueError
    if (y.ndim!=1):
        raise ValueError
    if (z.ndim!=1):
        raise ValueError
    
    ## Check if constant Δz (calculate Δz+ later)
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-7)):
        raise NotImplementedError
    
    ## Check if mean rgd has attr 'dt'
    if ('dt' in self.attrs.keys()):
        dt = self.attrs['dt']
        if (dt is not None):
            dt *= self.tchar
    else:
        raise ValueError
    
    if verbose: even_print('Δt/tchar','%0.8f'%(dt/self.tchar))
    if verbose: even_print('Δt','%0.3e [s]'%(dt,))
    if verbose: even_print('duration/tchar','%0.1f'%(self.duration_avg,))
    if verbose: even_print('duration','%0.3e [s]'%(self.duration_avg*self.tchar,))
    if verbose: print(72*'-')
    
    ## Re-dimensionalize
    u   =  np.copy( self.U_inf                     *  self['data/u'][()].T   )
    v   =  np.copy( self.U_inf                     *  self['data/v'][()].T   )
    w   =  np.copy( self.U_inf                     *  self['data/w'][()].T   )
    rho =  np.copy( self.rho_inf                   *  self['data/rho'][()].T )
    p   =  np.copy( (self.rho_inf * self.U_inf**2) *  self['data/p'][()].T   )
    T   =  np.copy( self.T_inf                     *  self['data/T'][()].T   )
    
    # mu1 = np.copy( self.C_Suth * T**(3/2) / (T + self.S_Suth) )
    # mu2 = np.copy( self.mu_Suth_ref * ( T / self.T_Suth_ref )**(3/2) * ((self.T_Suth_ref+self.S_Suth)/(T+self.S_Suth)) )
    # np.testing.assert_allclose(mu1, mu2, rtol=2e-7, atol=2e-7)
    # mu = np.copy(mu2)
    
    mu = np.copy( self.C_Suth * T**(3/2) / (T + self.S_Suth) )
    nu = np.copy( mu / rho )
    
    # === Average in [x,z] --> leave 1D [y]
    
    u   = np.squeeze( np.mean( u   , axis=(0,2), dtype=np.float64) )
    v   = np.squeeze( np.mean( v   , axis=(0,2), dtype=np.float64) )
    w   = np.squeeze( np.mean( w   , axis=(0,2), dtype=np.float64) )
    rho = np.squeeze( np.mean( rho , axis=(0,2), dtype=np.float64) )
    p   = np.squeeze( np.mean( p   , axis=(0,2), dtype=np.float64) )
    T   = np.squeeze( np.mean( T   , axis=(0,2), dtype=np.float64) )
    mu  = np.squeeze( np.mean( mu  , axis=(0,2), dtype=np.float64) )
    nu  = np.squeeze( np.mean( nu  , axis=(0,2), dtype=np.float64) )
    
    # ## determine finite difference order / size of central stencil based on [nx]
    # if (nx<3):
    #     raise ValueError('dx[] not possible because nx<3')
    # elif (nx>=3) and (nx<5):
    #     acc = 2
    # elif (nx>=5) and (nx<7):
    #     acc = 4
    # elif (nx>=7):
    #     acc = 6
    # else:
    #     raise ValueError('this should never happen')
    
    if verbose: even_print('acc','%i'%acc)
    if verbose: even_print('edge_stencil',edge_stencil)
    if verbose: print(72*'-')
    
    ## Get [y] gradients --> size [y]
    ddy_u   = gradient(u   , y, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddy_v   = gradient(v   , y, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddy_T   = gradient(T   , y, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddy_p   = gradient(p   , y, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddy_rho = gradient(rho , y, axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    
    ## Wall quantities
    ddy_u_wall  = float( ddy_u[0] )
    ddy_T_wall  = float( ddy_T[0] )
    rho_wall    = float( rho[0]   )
    nu_wall     = float( nu[0]    )
    mu_wall     = float( mu[0]    )
    T_wall      = float( T[0]     )
    tau_wall    = mu_wall * ddy_u_wall
    
    u_tau   = np.sqrt( tau_wall / rho_wall ) ## uτ
    sc_u_in = u_tau
    sc_l_in = nu_wall / u_tau     ## δν
    sc_t_in = nu_wall / u_tau**2  ## tν
    np.testing.assert_allclose(sc_t_in, sc_l_in/sc_u_in, rtol=1e-14, atol=1e-14)
    
    if verbose: even_print('ρw',f'{rho_wall:0.5f} [kg/m³]')
    if verbose: even_print('Tw',f'{T_wall:0.2f} [K]')
    if verbose: even_print('μw',f'{mu_wall:0.5E} [kg/(m·s)]')
    if verbose: even_print('νw',f'{nu_wall:0.5E} [m²/s]')
    if verbose: even_print('uτ',f'{u_tau:0.5E} [m/s]')
    if verbose: even_print('τw',f'{tau_wall:0.5E} [Pa]')
    if verbose: even_print('δν',f'{sc_l_in:0.5E} [m]')
    if verbose: even_print('tν',f'{sc_t_in:0.5E} [s]')
    if verbose: even_print('(du/dy)_w',f'{ddy_u_wall:0.3E} [1/s]')
    if verbose: even_print('(dT/dy)_w',f'{ddy_T_wall:0.3E} [K/m]')
    if verbose: print(72*'-')
    
    ddy_u_plus = np.copy( ddy_u / (u_tau/sc_l_in) ) ## du+/dy+ = (du/dy)/(uτ/δν) = (du/dy)/(uτ^2/νw)
    
    ## Get edge --> where |du+/dy+|<ϵ
    y_edge,j_edge = calc_profile_edge_1d(
                        y=y, ## dimensional
                        ddy_u=ddy_u_plus,
                        epsilon=epsilon,
                        )
    
    if verbose: even_print('ϵ',f'{epsilon:0.5E}')
    if verbose: even_print('j_edge',f'{j_edge:d}')
    if verbose: even_print('(du+/dy+)_edge',f'{ddy_u_plus[j_edge]:0.5E}')
    if verbose: even_print('y_edge',f'{y_edge:0.5E} [m]')
    if verbose: even_print('y+_edge',f'{y_edge/sc_l_in:0.3f}')
    
    aa = (y_edge-y.min())/(y.max()-y.min())
    if verbose: even_print('(y_edge-ymin)/(ymax-ymin)',f'{aa:0.3f}')
    
    # ===
    
    u_edge   =   u[j_edge]
    #v_edge   =   v[j_edge]
    #w_edge   =   w[j_edge]
    T_edge   =   T[j_edge]
    rho_edge = rho[j_edge]
    mu_edge  =  mu[j_edge]
    nu_edge  =  nu[j_edge]
    
    ## δ99
    d99 = calc_d99_1d(
            y=y,
            u=u,
            j_edge=j_edge,
            u_edge=u_edge,
            rtol=1e-3,
            interp_kind='cubic',
            )
    
    u99 = float( sp.interpolate.interp1d(y, u, kind='cubic', bounds_error=True)(d99) )
    
    ## Outer scales
    sc_l_out = d99
    sc_u_out = u99
    sc_t_out = d99/u99
    np.testing.assert_allclose(sc_t_out, sc_l_out/sc_u_out, rtol=1e-14, atol=1e-14)
    
    t_meas = self.duration_avg * (self.lchar / self.U_inf)
    t_eddy = t_meas / ( d99 / u_tau )
    
    ## Get BL integral quantities
    dd = calc_bl_integral_quantities_1d(
            y=y,
            u=u,
            rho=rho,
            d99=d99,
            u_tau=u_tau, rho_wall=rho_wall, mu_wall=mu_wall,
            u_edge=u_edge, rho_edge=rho_edge, mu_edge=mu_edge,
            j_edge=j_edge,
            )
    
    # === Add to file
    
    gn = 'data_dim'
    
    ## if group already exists in file, delete it entirely
    if (gn in self):
        del self[gn]
    
    ## 1D
    
    self.create_dataset(f'{gn}/u'     , data=u     , chunks=None)
    self.create_dataset(f'{gn}/rho'   , data=rho   , chunks=None)
    
    self.create_dataset(f'{gn}/ddy_u'   , data=ddy_u   , chunks=None)
    self.create_dataset(f'{gn}/ddy_v'   , data=ddy_v   , chunks=None)
    self.create_dataset(f'{gn}/ddy_T'   , data=ddy_T   , chunks=None)
    self.create_dataset(f'{gn}/ddy_p'   , data=ddy_p   , chunks=None)
    self.create_dataset(f'{gn}/ddy_rho' , data=ddy_rho , chunks=None)
    
    self.create_dataset(f'{gn}/z1d' , data=z , chunks=None)
    self.create_dataset(f'{gn}/z'   , data=z , chunks=None)
    
    ## 0D
    
    self.create_dataset(f'{gn}/dz0'        , data=dz0 , chunks=None)
    self.create_dataset(f'{gn}/dt'         , data=dt  , chunks=None)
    
    self.create_dataset(f'{gn}/y_edge'     , data=y_edge    , chunks=None)
    self.create_dataset(f'{gn}/d99'        , data=d99       , chunks=None)
    self.create_dataset(f'{gn}/u99'       , data=u99      , chunks=None)
    
    self.create_dataset(f'{gn}/u_edge'     , data=u_edge     , chunks=None)
    self.create_dataset(f'{gn}/rho_edge'   , data=rho_edge   , chunks=None)
    self.create_dataset(f'{gn}/mu_edge'    , data=mu_edge    , chunks=None)
    self.create_dataset(f'{gn}/nu_edge'    , data=nu_edge    , chunks=None)
    self.create_dataset(f'{gn}/T_edge'     , data=T_edge     , chunks=None)
    
    self.create_dataset(f'{gn}/u_tau'      , data=u_tau      , chunks=None)
    
    self.create_dataset(f'{gn}/ddy_u_wall' , data=ddy_u_wall , chunks=None )
    self.create_dataset(f'{gn}/ddy_T_wall' , data=ddy_T_wall , chunks=None )
    self.create_dataset(f'{gn}/rho_wall'   , data=rho_wall   , chunks=None )
    self.create_dataset(f'{gn}/nu_wall'    , data=nu_wall    , chunks=None )
    self.create_dataset(f'{gn}/mu_wall'    , data=mu_wall    , chunks=None )
    self.create_dataset(f'{gn}/T_wall'     , data=T_wall     , chunks=None )
    self.create_dataset(f'{gn}/tau_wall'   , data=tau_wall   , chunks=None )
    #self.create_dataset(f'{gn}/q_wall'     , data=q_wall     , chunks=None )
    
    self.create_dataset(f'{gn}/sc_u_in'    , data=sc_u_in    , chunks=None)
    self.create_dataset(f'{gn}/sc_l_in'    , data=sc_l_in    , chunks=None)
    self.create_dataset(f'{gn}/sc_t_in'    , data=sc_t_in    , chunks=None)
    self.create_dataset(f'{gn}/sc_u_out'   , data=sc_u_out   , chunks=None)
    self.create_dataset(f'{gn}/sc_l_out'   , data=sc_l_out   , chunks=None)
    self.create_dataset(f'{gn}/sc_t_out'   , data=sc_t_out   , chunks=None)
    
    ## Add integrated quantities (all 0D)
    for key,val in dd.items():
        self.create_dataset(f'{gn}/{key}', data=val, chunks=None)
    
    ## Report -- currently does not report everything returned
    if verbose:
        print(72*'-')
        even_print('Reτ'    , '%0.1f'%dd['Re_tau']      )
        even_print('Reθ'    , '%0.1f'%dd['Re_theta']    )
        even_print('Reδ*'   , '%0.1f'%dd['Re_dstar']    )
        even_print('Reδ2'   , '%0.1f'%dd['Re_d2']       )
        even_print('θ'      , '%0.5E [m]'%dd['theta']   )
        even_print('δ*'     , '%0.5E [m]'%dd['dstar']   )
        even_print('Δ'      , '%0.5E [m]'%dd['dRC']     )
        even_print('H12'    , '%0.5f'%dd['H12']         )
        even_print('H32'    , '%0.5f'%dd['H32']         )
        even_print('δ99'    , '%0.5E [m]'%d99           )
        even_print('θ/δ99'  , '%0.5f'%(dd['theta']/d99) )
        even_print('δ*/δ99' , '%0.5f'%(dd['dstar']/d99) )
        
        even_print('uτ'     , f'{u_tau:0.5E} [m/s]'        )
        even_print('νw'     , f'{nu_wall:0.5E} [m²/s]'     )
        even_print('μw'     , f'{mu_wall:0.5E} [kg/(m·s)]' )
        even_print('τw'     , f'{tau_wall:0.5E} [Pa]'      )
        
        even_print('τw/q_inf'                 , '%0.5E'%(tau_wall/(self.rho_inf*self.U_inf**2)) )
        even_print('cf = 2·τw/(ρe·ue²)'       , '%0.5E'%(2*tau_wall/(rho_edge*u_edge**2)) )
        even_print('t_meas'                   , '%0.5E [s]'%t_meas            )
        even_print('t_meas/tchar'             , '%0.1f'%(t_meas/self.tchar)   )
        even_print('t_eddy = t_meas/(δ99/uτ)' , '%0.2f'%t_eddy                )
        even_print('t_meas/(δ99/u99)'         , '%0.2f'%(t_meas/(d99/u99))    )
        even_print('t_meas/(20·δ99/u99)'      , '%0.2f'%(t_meas/(20*d99/u99)) )
        print(72*'-')
        even_print('uτ'           , '%0.5E [m/s]'%(sc_u_in,)  )
        even_print('δν = νw/uτ'   , '%0.5E [m]'%(sc_l_in,)    )
        even_print('tν = νw/uτ²'  , '%0.5E [s]'%(sc_t_in,)    )
        even_print('u99'          , '%0.5E [m/s]'%(sc_u_out,) )
        even_print('δ99'          , '%0.5E [m]'%(sc_l_out,)   )
        even_print('δ99/u99'      , '%0.5E [s]'%(sc_t_out,)   )
    
    if verbose: print(72*'-')
    if verbose: print('total time : rgd.add_mean_dimensional_data_xpln() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import sys
import timeit

import numpy as np
from tqdm import tqdm

from turbx import gradient
from turbx.grid_metric import get_metric_tensor_2d

from .utils import even_print, format_time_string

'''
========================================================================
Mass & Momentum Balance Tools
========================================================================
'''

# ======================================================================
# Mass
# ======================================================================

@staticmethod
def mass_balance_kernel_cartesian(x,y, u_Fv,v_Fv,rho, transformation_matrix=None, acc=6, edge_stencil='full'):
    '''
    Get Cartesian [x,y] mass balance terms in 2D plane
    '''
    
    nx,ny = x.shape
    
    ## Rotate & translate [x,y] and rotate vector quantities (global Cartesian)
    if transformation_matrix is not None:
        
        ## Rotate & translate [x,y]
        xy2d_tmp = np.stack((x, y, np.ones_like(x)), axis=-1)
        xy2d_tmp = np.einsum( 'ij,xyj->xyi', transformation_matrix , xy2d_tmp )
        x        = xy2d_tmp[:,:,0]
        y        = xy2d_tmp[:,:,1]
        xy2d_tmp = None; del xy2d_tmp
        
        ## Make a copy of the [2x2] matrix
        rotation_matrix_2x2 = np.copy(transformation_matrix[:2,:2])
        
        ## Rotate [u_Fv,v_Fv]
        uv2d_Fv = np.stack((u_Fv,v_Fv), axis=-1)
        uv2d_Fv = np.einsum('ij,xyj->xyi', rotation_matrix_2x2, uv2d_Fv)
        u_Fv    = uv2d_Fv[:,:,0]
        v_Fv    = uv2d_Fv[:,:,1]
        uv2d_Fv = None; del uv2d_Fv
    
    # ==================================================================
    # Get metric tensor
    # M = δ[q1,q2]/δ[x1,x2] : δ[q1,q2]/δ[x,y] or δ[q1,q2]/δ[r,θ]
    # ==================================================================
    
    M = get_metric_tensor_2d(x, y, acc=acc, edge_stencil=edge_stencil, verbose=False)
    ddx1_q1 = M[:,:,0,0] ## ∂q1/∂x1 --> ∂q1/∂x or ∂q1/∂r
    ddx1_q2 = M[:,:,1,0] ## ∂q2/∂x1 --> ∂q2/∂x or ∂q2/∂r
    ddx2_q1 = M[:,:,0,1] ## ∂q1/∂x2 --> ∂q1/∂y or ∂q1/∂θ
    ddx2_q2 = M[:,:,1,1] ## ∂q2/∂x2 --> ∂q2/∂y or ∂q2/∂θ
    M = None ; del M
    
    # ==================================================================
    # Construct mass balance : Cartesian [x,y]
    # ==================================================================
    
    # === ∂x[ρ·u]
    
    A      = rho * u_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t1 = ddx1_A
    
    # === ∂y[ρ·v]
    
    A      = rho * v_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t2 = ddx2_A
    
    # ===
    
    data = {
        't1':t1,
        't2':t2,
        }
    
    return data

@staticmethod
def mass_balance_kernel_cylindrical(x,y, u_Fv,v_Fv,rho, transformation_matrix=None, acc=6, edge_stencil='full'):
    '''
    Get cylindrical [r,θ] mass balance terms in 2D plane
    '''
    
    nx,ny = x.shape
    
    ## Rotate & translate [x,y] and rotate vector quantities (global Cartesian)
    if transformation_matrix is not None:
        
        ## Rotate & translate [x,y]
        xy2d_tmp = np.stack((x, y, np.ones_like(x)), axis=-1)
        xy2d_tmp = np.einsum( 'ij,xyj->xyi', transformation_matrix , xy2d_tmp )
        x        = xy2d_tmp[:,:,0]
        y        = xy2d_tmp[:,:,1]
        xy2d_tmp = None; del xy2d_tmp
        
        ## Make a copy of the [2x2] matrix
        rotation_matrix_2x2 = np.copy(transformation_matrix[:2,:2])
        
        ## Rotate [u_Fv,v_Fv]
        uv2d_Fv = np.stack((u_Fv,v_Fv), axis=-1)
        uv2d_Fv = np.einsum('ij,xyj->xyi', rotation_matrix_2x2, uv2d_Fv)
        u_Fv    = uv2d_Fv[:,:,0]
        v_Fv    = uv2d_Fv[:,:,1]
        uv2d_Fv = None; del uv2d_Fv
    
    ## Calculate coordinate arrays [r,θ] from [x,y]
    r        = np.sqrt(x**2 + y**2)
    theta    = np.arctan2(y,x)
    costheta = np.cos(theta)
    sintheta = np.sin(theta)
    
    eps = np.finfo(np.float64).eps
    if np.any(r<eps):
        raise RuntimeError(f'r has values < {eps:0.3e}')
    
    ## Calculate [ur_Fv,uθ_Fv] from [u_Fv,v_Fv]
    ur_Fv =  u_Fv*costheta + v_Fv*sintheta
    ut_Fv = -u_Fv*sintheta + v_Fv*costheta
    
    # ==================================================================
    # Get metric tensor
    # M = δ[q1,q2]/δ[x1,x2] : δ[q1,q2]/δ[x,y] or δ[q1,q2]/δ[r,θ]
    # ==================================================================
    
    M = get_metric_tensor_2d(r, theta, acc=acc, edge_stencil=edge_stencil, verbose=False)
    ddx1_q1 = M[:,:,0,0] ## ∂q1/∂x1 --> ∂q1/∂x or ∂q1/∂r
    ddx1_q2 = M[:,:,1,0] ## ∂q2/∂x1 --> ∂q2/∂x or ∂q2/∂r
    ddx2_q1 = M[:,:,0,1] ## ∂q1/∂x2 --> ∂q1/∂y or ∂q1/∂θ
    ddx2_q2 = M[:,:,1,1] ## ∂q2/∂x2 --> ∂q2/∂y or ∂q2/∂θ
    M = None ; del M
    
    # ==================================================================
    # Construct mass balance : cylindrical [r,θ]
    # ==================================================================
    
    ## ∇·(ρu) = ∂r[ρ·ur] + (1/r)∂θ[ρ·uθ] + (1/r)[ρ·ur] = 0
    
    # === ∂r[ρ·ur]
    
    A = rho * ur_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t1 = ddx1_A
    
    # === (1/r)·∂θ[ρ·uθ]
    
    A = rho * ut_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t2 = (1/r) * ddx2_A
    
    # === (1/r)·[ρ·ur]
    
    t3 = (1/r) * rho * ur_Fv
    
    # ===
    
    data = {
        't1':t1,
        't2':t2,
        't3':t3,
        }
    
    return data

def _calc_mass_balance_terms(self, **kwargs):
    '''
    Calculate terms of time-averaged mass conservation equation
    '''
    
    verbose      = kwargs.get('verbose',True)
    local_csys   = kwargs.get('local_csys',True)
    acc          = kwargs.get('acc',6)
    edge_stencil = kwargs.get('edge_stencil','full')
    xis          = kwargs.get('xis',None) ## which [x] indices to compute (None=all)
    csys_type    = kwargs.get('csys_type','cartesian')
    
    if verbose: print('\n'+'ztmd.calc_mass_balance_terms()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if self.open_mode not in ['a','w','r+']:
        raise ValueError('Not able to write to hdf5 file')
    
    if verbose: even_print('local_csys',str(local_csys))
    if verbose: even_print('csys_type',csys_type)
    
    if local_csys:
        if self.rectilinear:
            print('>>> WARNING: this grid is rectilinear. Using local csys is not needed.')
        if 'csys/vtang' not in self:
            raise ValueError('csys/vtang not in hdf5')
        if 'csys/vnorm' not in self:
            raise ValueError('csys/vnorm not in hdf5')
        
        ## Read unit vectors (wall tangent, wall norm) from HDF5
        vtang = np.copy( self['csys/vtang'][()] )
        vnorm = np.copy( self['csys/vnorm'][()] )
        if verbose: even_print('vtang',str(vtang.shape))
        if verbose: even_print('vnorm',str(vnorm.shape))
    
    if (csys_type=='cartesian'):
        gn = 'data_mass_xy'
        gns = [gn,]
    elif (csys_type=='cylindrical'):
        gnr = 'data_mass_rtheta'  ## keep single group for mass in cyl
        gns = [gnr,]
    else:
        raise NotImplementedError
    
    if (csys_type=='cartesian'):
        terms = {
            't1':(r'∂x[ρ·u]', r'$\partial_{x}[\bar{\rho} \tilde{u}]$'),
            't2':(r'∂y[ρ·v]', r'$\partial_{y}[\bar{\rho} \tilde{v}]$'),
            }
    elif (csys_type=='cylindrical'):
        terms = {
            't1':(r'∂r[ρ·ur]'       , r'$\partial_{r}[ \bar{\rho} \tilde{u}_r ]$'),
            't2':(r'(1/r)·∂θ[ρ·uθ]' , r'$\frac{1}{r}\partial_{\theta}[ \bar{\rho} \tilde{u}_{\theta} ]$'),
            't3':(r'(1/r)·[ρ·ur]'   , r'$\frac{1}{r}[ \bar{\rho} \tilde{u}_r ]$'),
            }
    else:
        raise ValueError
    
    ## Create groups in ZTMD
    for gn in gns:
        if verbose: print(72*'-')
        if gn in self:
            del self[gn]
            if verbose: even_print('deleted group',f'/{gn}/')
        
        grp = self.create_group(gn)
        if verbose: even_print('created group',f'{grp.name}/')
        
        ## Set group attributes
        grp.attrs['local_csys'] = local_csys
        if verbose: even_print(f'{gn}.local_csys',str(grp.attrs['local_csys']))
    
    gn=None; del gn
    
    if verbose: print(72*'-')
    
    ## Initialize dsets (Cartesian)
    if (csys_type=='cartesian'):
        for dsn, tags in terms.items():
            unicode,latex = tags
            shape = (self.ny,self.nx)
            if (f'{gns[0]}/{dsn}' in self):
                del self[f'{gns[0]}/{dsn}']
            dset = self.create_dataset(
                                    f'{gns[0]}/{dsn}', 
                                    shape=shape, 
                                    dtype=np.float32,
                                    chunks=None,
                                    fillvalue=np.float32(np.nan),
                                    )
            if verbose: even_print(f'{gns[0]}/{dsn}',str(shape))
            dset.attrs['latex'] = latex
            dset.attrs['unicode'] = unicode
        if verbose: print(72*'-')
    
    ## Initialize dsets (cylindrical)
    if (csys_type=='cylindrical'):
        for dsn, tags in terms.items():
            unicode,latex = tags
            shape = (self.ny,self.nx)
            if (f'{gnr}/{dsn}' in self):
                del self[f'{gnr}/{dsn}']
            dset = self.create_dataset(
                                    f'{gnr}/{dsn}', 
                                    shape=shape, 
                                    dtype=np.float32,
                                    chunks=None,
                                    fillvalue=np.float32(np.nan),
                                    )
            if verbose: even_print(f'{gnr}/{dsn}',str(shape))
            dset.attrs['latex'] = latex
            dset.attrs['unicode'] = unicode
        if verbose: print(72*'-')
    
    ## Read data
    if csys_type=='cylindrical' and hasattr(self,'crv_R'):
        crv_R = np.copy(self.crv_R)
        if verbose:
            even_print('crv_R',str(crv_R.shape))
            print(72*'-')
    
    if verbose:
        print('Reading data'+'\n'+'------------')
    #u   = np.copy( self['data/u'][()].T   )
    #v   = np.copy( self['data/v'][()].T   )
    rho = np.copy( self['data/rho'][()].T )
    
    #if verbose: even_print('u',str(u.shape))
    #if verbose: even_print('v',str(v.shape))
    if verbose: even_print('ρ',str(rho.shape))
    
    u_Fv = np.copy( self['data/u_Fv'][()].T )
    v_Fv = np.copy( self['data/v_Fv'][()].T )
    if verbose: even_print('u_Fv',str(u_Fv.shape))
    if verbose: even_print('v_Fv',str(v_Fv.shape))
    
    x = np.copy(self.x)
    y = np.copy(self.y)
    
    if x.ndim==1 and y.ndim==1:
        x,y = np.meshgrid(x,y, indexing='ij')
    if verbose: even_print('x',str(x.shape))
    if verbose: even_print('y',str(y.shape))
    if verbose: print(72*'-')
    
    nx,ny = x.shape
    
    # ==================================================================
    
    def _get_transformation_matrix(xi):
        '''
        Get transformation matrix : translate,rotate,[translate]
        '''
        rotation_angle_z = np.arcsin(vnorm[xi,0,0])
        
        rotation_matrix = np.array( [[ np.cos(rotation_angle_z) , -np.sin(rotation_angle_z) , 0 ],
                                     [ np.sin(rotation_angle_z) ,  np.cos(rotation_angle_z) , 0 ],
                                     [ 0                        ,  0                        , 1 ]], dtype=np.float64 )
        
        ## Translate to [cx,cy]
        cx = x[xi,0]
        cy = y[xi,0]
        translation_matrix_A = np.array( [[ 1 , 0 , -cx ],
                                          [ 0 , 1 , -cy ],
                                          [ 0 , 0 ,  1  ]], dtype=np.float64 )
        
        ## For cylindrical csys, translate in [y] by -R (the local curvature radius)
        ##   such that origin is distance 'R' from the wall
        if (csys_type=='cylindrical'):
            cx = 0.
            try:
                cy = crv_R[xi]
            except NameError:
                cy = 0.
            translation_matrix_B = np.array( [[ 1 , 0 ,  0  ],
                                              [ 0 , 1 , -cy ],
                                              [ 0 , 0 ,  1  ]], dtype=np.float64 )
        else:
            translation_matrix_B = np.identity(3, dtype=np.float64) ## No second translation
        
        ## Form the full [3x3] matrix by multiplying together
        ##  --> ORDER MATTERS!
        transformation_matrix = np.einsum( 'ij,jk,kl->il', translation_matrix_B , rotation_matrix, translation_matrix_A )
        
        return transformation_matrix
    
    # ==================================================================
    
    if local_csys: ## Compute mass balance in local csys
        
        if xis is None: ## Do all
            xis = np.arange(self.nx)
            n_prog = self.nx
        else: ## Do only for selected x-indices
            n_prog = len(xis)
        
        ## Assert passed 'xis' is OK
        if isinstance(xis,np.ndarray): ## numpy array
            if not xis.ndim==1:
                raise ValueError('xis must have .ndim==1')
            if not np.issubdtype(xis.dtype,np.integer):
                raise RuntimeError("xis must have subtype np.integer")
        else: ## some other iterable... are all elements ints?
            try:
                xis = [int(xi) for xi in xis]
            except Exception as e:
                raise RuntimeError("elements of xis are not interpretable as ints") from e
        
        # ==============================================================
        # Main loop
        # ==============================================================
        
        if csys_type=='cartesian':
            desc='Mass balance [x,y]'
        elif csys_type=='cylindrical':
            desc='Mass balance [r,θ]'
        else:
            raise ValueError
        
        if verbose:
            progress_bar = tqdm(
                total=n_prog,
                ncols=100,
                desc=desc,
                leave=True,
                file=sys.stdout,
                mininterval=0.1,
                smoothing=0.,
                bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
                ascii="░█",
                colour='#00ffff',
                )
        
        for xi in xis:
            
            ## Get transformation matrix (global-->local) for this [x] index
            transformation_matrix = _get_transformation_matrix(xi)
            
            ## Take N points in [x] on either side
            ## This allows for full FD stencil at central point
            xiA = max( 0  , xi-12   )
            xiB = min( nx , xi+12+1 )
            xii = xi - xiA ## Index local to slice
            
            if (csys_type=='cartesian'):
                
                data = mass_balance_kernel_cartesian(
                                                x=x[xiA:xiB,:],
                                                y=y[xiA:xiB,:],
                                                u_Fv=u_Fv[xiA:xiB,:],
                                                v_Fv=v_Fv[xiA:xiB,:],
                                                rho=rho[xiA:xiB,:],
                                                transformation_matrix=transformation_matrix,
                                                acc=acc,
                                                edge_stencil=edge_stencil,
                                                )
                
                ## Write to HDF5
                for dsn, arr in data.items():
                    dset = self[f'{gns[0]}/{dsn}']
                    dset[:,xi] = arr[xii,:] ## Write only 1 [x] position
            
            elif (csys_type=='cylindrical'):
                
                data = mass_balance_kernel_cylindrical(
                                                x=x[xiA:xiB,:],
                                                y=y[xiA:xiB,:],
                                                u_Fv=u_Fv[xiA:xiB,:],
                                                v_Fv=v_Fv[xiA:xiB,:],
                                                rho=rho[xiA:xiB,:],
                                                transformation_matrix=transformation_matrix,
                                                acc=acc,
                                                edge_stencil=edge_stencil,
                                                )
                
                ## Write to HDF5
                for dsn, arr in data.items():
                    dset = self[f'{gnr}/{dsn}']
                    dset[:,xi] = arr[xii,:] ## Write only 1 [x] position
            
            else:
                raise ValueError
            
            if verbose: progress_bar.update()
        if verbose: progress_bar.close()
    
    else: ## Compute mass balance in global csys
        
        ## Send to kernel
        if csys_type=='cartesian':
            
            data = mass_balance_kernel_cartesian(
                                            x=x,
                                            y=y,
                                            u_Fv=u_Fv,
                                            v_Fv=v_Fv,
                                            rho=rho,
                                            #transformation_matrix=transformation_matrix,
                                            acc=acc,
                                            edge_stencil=edge_stencil,
                                            )
            
            ## Write to HDF5
            for dsn, arr in data.items():
                dset = self[f'{gns[0]}/{dsn}']
                if verbose:
                    even_print(dset.name,str(arr.T.shape))
                dset[:,:] = arr.T
        
        elif csys_type=='cylindrical':
            
            data = mass_balance_kernel_cylindrical(
                                            x=x,
                                            y=y,
                                            u_Fv=u_Fv,
                                            v_Fv=v_Fv,
                                            rho=rho,
                                            #transformation_matrix=transformation_matrix,
                                            acc=acc,
                                            edge_stencil=edge_stencil,
                                            )
            
            ## Write to HDF5
            for dsn, arr in data.items():
                dset = self[f'{gnr}/{dsn}']
                if verbose:
                    even_print(dset.name,str(arr.T.shape))
                dset[:,:] = arr.T
        
        else:
            raise ValueError
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_mass_balance_terms() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return

# ======================================================================
# Momentum
# ======================================================================

@staticmethod
def momentum_balance_kernel_cartesian(x,y, u,v, u_Fv,v_Fv, p,rho,mu, r_uII_uII,r_uII_vII,r_vII_vII, transformation_matrix=None, acc=6, edge_stencil='full'):
    '''
    Get Cartesian [x,y] momentum balance terms in 2D plane
    '''
    
    nx,ny = x.shape
    
    ## Rotate & translate [x,y] and rotate vector/tensor quantities (global Cartesian)
    if transformation_matrix is not None:
        
        ## Rotate & translate [x,y]
        xy2d_tmp = np.stack((x, y, np.ones_like(x)), axis=-1)
        xy2d_tmp = np.einsum( 'ij,xyj->xyi', transformation_matrix , xy2d_tmp )
        x        = xy2d_tmp[:,:,0]
        y        = xy2d_tmp[:,:,1]
        xy2d_tmp = None; del xy2d_tmp
        
        ## Make a copy of the [2x2] matrix
        rotation_matrix_2x2 = np.copy(transformation_matrix[:2,:2])
        
        ## Rotate [u,v]
        uv2d = np.stack((u,v), axis=-1)
        uv2d = np.einsum('ij,xyj->xyi' , rotation_matrix_2x2 , uv2d)
        u    = uv2d[:,:,0]
        v    = uv2d[:,:,1]
        uv2d = None; del uv2d
        
        ## Rotate [u_Fv,v_Fv]
        uv2d_Fv = np.stack((u_Fv,v_Fv), axis=-1)
        uv2d_Fv = np.einsum('ij,xyj->xyi', rotation_matrix_2x2, uv2d_Fv)
        u_Fv    = uv2d_Fv[:,:,0]
        v_Fv    = uv2d_Fv[:,:,1]
        uv2d_Fv = None; del uv2d_Fv
        
        # === Turbulent stress tensor: global Cartesian rotate
        
        r_uiIIujII = np.zeros((nx,ny,2,2), dtype=np.float64)
        r_uiIIujII[:,:,0,0] = r_uII_uII
        r_uiIIujII[:,:,0,1] = r_uII_vII
        r_uiIIujII[:,:,1,0] = r_uII_vII
        r_uiIIujII[:,:,1,1] = r_vII_vII
        
        ## Assert symmetry
        np.testing.assert_allclose(r_uiIIujII, np.transpose(r_uiIIujII, (0,1,3,2)), atol=1e-8, rtol=1e-8)
        
        ## Tensor rotation : [A'] = [R][A][R]^T
        r_uiIIujII = np.copy( np.einsum( 'ij,xyjk,kl->xyil' , rotation_matrix_2x2 , r_uiIIujII , rotation_matrix_2x2.T ) )
        
        r_uII_uII = r_uiIIujII[:,:,0,0]
        r_uII_vII = r_uiIIujII[:,:,0,1]
        r_vII_uII = r_uiIIujII[:,:,1,0]
        r_vII_vII = r_uiIIujII[:,:,1,1]
        
        r_uiIIujII = None ; del r_uiIIujII
        
        np.testing.assert_allclose(r_uII_vII, r_vII_uII, atol=1e-7, rtol=1e-7)
        r_vII_uII = None ; del r_vII_uII
    
    # ==================================================================
    # Get metric tensor
    # M = δ[q1,q2]/δ[x1,x2] : δ[q1,q2]/δ[x,y] or δ[q1,q2]/δ[r,θ]
    # ==================================================================
    
    M = get_metric_tensor_2d(x, y, acc=acc, edge_stencil=edge_stencil, verbose=False)
    ddx1_q1 = np.copy( M[:,:,0,0] ) ## ∂q1/∂x1 --> ∂q1/∂x or ∂q1/∂r
    ddx1_q2 = np.copy( M[:,:,1,0] ) ## ∂q2/∂x1 --> ∂q2/∂x or ∂q2/∂r
    ddx2_q1 = np.copy( M[:,:,0,1] ) ## ∂q1/∂x2 --> ∂q1/∂y or ∂q1/∂θ
    ddx2_q2 = np.copy( M[:,:,1,1] ) ## ∂q2/∂x2 --> ∂q2/∂y or ∂q2/∂θ
    M = None ; del M
    
    # ==================================================================
    # Calculate gradients : Cartesian [x,y]
    # ==================================================================
    
    ## [x1,x2] = [x,y]
    ## ∂(A)/∂x1 = ∂(A)/∂(q1)·∂(q1)/∂(x1) + ∂(A)/∂(q2)·∂(q2)/∂(x1)
    ## ∂(A)/∂x2 = ∂(A)/∂(q1)·∂(q1)/∂(x2) + ∂(A)/∂(q2)·∂(q2)/∂(x2)
    
    A = u
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_u  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_u  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    A = v
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_v  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_v  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    A = p
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_p  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_p  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    # ==================================================================
    # Construct momentum balance : Cartesian [x,y]
    # ==================================================================
    
    ## τ_xx, τ_yy, τ_xy
    tau_xx = +(4/3)*mu*ddx_u - (2/3)*mu*ddy_v
    tau_yy = -(2/3)*mu*ddx_u + (4/3)*mu*ddy_v
    tau_xy = mu*(ddx_v + ddy_u)
    
    # === ∂x[ρ·u·u]
    
    A      = rho * u_Fv * u_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t1x = ddx_A
    
    # === ∂x[ρ·u·v] & ∂y[ρ·u·v]
    
    A      = rho * u_Fv * v_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t1y = ddx_A
    t2x = ddy_A
    
    # === ∂y[ρ·v·v]
    
    A      = rho * v_Fv * v_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t2y = ddy_A
    
    # === ∂x[ρu″u″]
    
    A      = r_uII_uII
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t5x = ddx_A
    
    # === ∂x[ρu″v″] & ∂y[ρu″v″]
    
    A      = r_uII_vII
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t5y = ddx_A
    t7x = ddy_A
    
    # === ∂y[ρv″v″]
    
    A      = r_vII_vII
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t7y = ddy_A
    
    # === ∂x[p] & ∂y[p]
    
    t3x = ddx_p
    t3y = ddy_p
    
    # === -∂x[τxx]
    
    ddq1_A = gradient(tau_xx, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(tau_xx, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t4x = -1 * ddx_A
    
    # === -∂x[τxy] & -∂y[τxy]
    
    ddq1_A = gradient(tau_xy, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(tau_xy, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t4y = -1 * ddx_A
    t6x = -1 * ddy_A
    
    # === -∂y[τyy]
    
    ddq1_A = gradient(tau_yy, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(tau_yy, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx_A  = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddy_A  = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t6y = -1 * ddy_A
    
    # ===
    
    data_x = {
        't1':t1x,
        't2':t2x,
        't3':t3x,
        't4':t4x,
        't5':t5x,
        't6':t6x,
        't7':t7x,
        }
    
    data_y = {
        't1':t1y,
        't2':t2y,
        't3':t3y,
        't4':t4y,
        't5':t5y,
        't6':t6y,
        't7':t7y,
        }
    
    return data_x, data_y

@staticmethod
def momentum_balance_kernel_cylindrical(x,y, u,v, u_Fv,v_Fv, p,rho,mu, r_uII_uII,r_uII_vII,r_vII_vII, transformation_matrix=None, acc=6, edge_stencil='full'):
    '''
    Get cylindrical [r,θ] momentum balance terms in 2D plane
    '''
    
    nx,ny = x.shape
    
    ## Rotate & translate [x,y] and rotate vector/tensor quantities (global Cartesian)
    if transformation_matrix is not None:
        
        ## Rotate & translate [x,y]
        xy2d_tmp = np.stack((x, y, np.ones_like(x)), axis=-1)
        xy2d_tmp = np.einsum( 'ij,xyj->xyi', transformation_matrix , xy2d_tmp )
        x        = xy2d_tmp[:,:,0]
        y        = xy2d_tmp[:,:,1]
        xy2d_tmp = None; del xy2d_tmp
        
        ## Make a copy of the [2x2] matrix
        rotation_matrix_2x2 = np.copy(transformation_matrix[:2,:2])
        
        ## Rotate [u,v]
        uv2d = np.stack((u,v), axis=-1)
        uv2d = np.einsum('ij,xyj->xyi' , rotation_matrix_2x2 , uv2d)
        u    = uv2d[:,:,0]
        v    = uv2d[:,:,1]
        uv2d = None; del uv2d
        
        ## Rotate [u_Fv,v_Fv]
        uv2d_Fv = np.stack((u_Fv,v_Fv), axis=-1)
        uv2d_Fv = np.einsum('ij,xyj->xyi', rotation_matrix_2x2, uv2d_Fv)
        u_Fv    = uv2d_Fv[:,:,0]
        v_Fv    = uv2d_Fv[:,:,1]
        uv2d_Fv = None; del uv2d_Fv
        
        # === Turbulent stress tensor: global Cartesian rotate
        
        # a = rotation_matrix_2x2[0,0]
        # b = rotation_matrix_2x2[0,1]
        # c = rotation_matrix_2x2[1,0]
        # d = rotation_matrix_2x2[1,1]
        # 
        # r_uII_uII_new = a*a*r_uII_uII + 2*a*b*r_uII_vII + b*b*r_vII_vII
        # r_uII_vII_new = a*c*r_uII_uII + (a*d + b*c)*r_uII_vII + b*d*r_vII_vII
        # r_vII_vII_new = c*c*r_uII_uII + 2*c*d*r_uII_vII + d*d*r_vII_vII
        # 
        # r_uII_uII = r_uII_uII_new
        # r_uII_vII = r_uII_vII_new
        # r_vII_vII = r_vII_vII_new
        # 
        # r_uII_uII_new = None; del r_uII_uII_new
        # r_uII_vII_new = None; del r_uII_vII_new
        # r_vII_vII_new = None; del r_vII_vII_new
        
        # === Turbulent stress tensor: global Cartesian rotate
        
        r_uiIIujII = np.zeros((nx,ny,2,2), dtype=np.float64)
        r_uiIIujII[:,:,0,0] = np.copy( r_uII_uII )
        r_uiIIujII[:,:,0,1] = np.copy( r_uII_vII )
        r_uiIIujII[:,:,1,0] = np.copy( r_uII_vII )
        r_uiIIujII[:,:,1,1] = np.copy( r_vII_vII )
        
        ## Assert symmetry
        np.testing.assert_allclose(r_uiIIujII, np.transpose(r_uiIIujII, (0,1,3,2)), atol=1e-8, rtol=1e-8)
        
        ## Tensor rotation : [A'] = [R][A][R]^T
        r_uiIIujII = np.copy( np.einsum( 'ij,xyjk,kl->xyil' , rotation_matrix_2x2 , r_uiIIujII , rotation_matrix_2x2.T ) )
        
        r_uII_uII = np.copy(r_uiIIujII[:,:,0,0])
        r_uII_vII = np.copy(r_uiIIujII[:,:,0,1])
        r_vII_uII = np.copy(r_uiIIujII[:,:,1,0])
        r_vII_vII = np.copy(r_uiIIujII[:,:,1,1])
        
        r_uiIIujII = None ; del r_uiIIujII
        
        np.testing.assert_allclose(r_uII_vII, r_vII_uII, atol=1e-7, rtol=1e-7)
        r_vII_uII = None ; del r_vII_uII
    
    ## Calculate coordinate arrays [r,θ] from [x,y]
    r        = np.sqrt(x**2 + y**2)
    theta    = np.arctan2(y,x)
    costheta = np.cos(theta)
    sintheta = np.sin(theta)
    
    eps = np.finfo(np.float64).eps
    if np.any(r<eps):
        raise RuntimeError(f'r has values < {eps:0.3e}')
    
    ## Calculate [ur,uθ] from [u,v]
    ur =  u*costheta + v*sintheta
    ut = -u*sintheta + v*costheta
    
    ## Calculate [ur_Fv,uθ_Fv] from [u_Fv,v_Fv]
    ur_Fv =  u_Fv*costheta + v_Fv*sintheta
    ut_Fv = -u_Fv*sintheta + v_Fv*costheta
    
    ## Re-Stress tensor : convert from Cartesian (rotated) to cylindrical components
    ## Tensor transform for symmetric 2x2: [R][T][R]^T, with R = [[c,s],[-s,c]]
    ## Components:
    ## rr = c^2*uu + 2cs*uv + s^2*vv
    ## rt = -cs*uu + (c^2 - s^2)*uv + cs*vv
    ## tt = s^2*uu - 2cs*uv + c^2*vv
    
    c2  = costheta*costheta
    s2  = sintheta*sintheta
    cs  = costheta*sintheta
    c2ms2 = c2 - s2
    
    r_urII_urII =  c2*r_uII_uII +  2*cs*r_uII_vII + s2*r_vII_vII
    r_urII_utII = -cs*r_uII_uII + c2ms2*r_uII_vII + cs*r_vII_vII
    r_utII_urII = r_urII_utII
    r_utII_utII = s2*r_uII_uII - 2*cs*r_uII_vII + c2*r_vII_vII
    
    ## Assert symmetry
    np.testing.assert_allclose(r_urII_utII, r_utII_urII, atol=1e-7, rtol=1e-7)
    r_utII_urII = None ; del r_utII_urII
    
    # ==================================================================
    # Get metric tensor
    # M = δ[q1,q2]/δ[x1,x2] : δ[q1,q2]/δ[x,y] or δ[q1,q2]/δ[r,θ]
    # ==================================================================
    
    M = get_metric_tensor_2d(r, theta, acc=acc, edge_stencil=edge_stencil, verbose=False)
    ddx1_q1 = M[:,:,0,0] ## ∂q1/∂x1 --> ∂q1/∂x or ∂q1/∂r
    ddx1_q2 = M[:,:,1,0] ## ∂q2/∂x1 --> ∂q2/∂x or ∂q2/∂r
    ddx2_q1 = M[:,:,0,1] ## ∂q1/∂x2 --> ∂q1/∂y or ∂q1/∂θ
    ddx2_q2 = M[:,:,1,1] ## ∂q2/∂x2 --> ∂q2/∂y or ∂q2/∂θ
    M = None ; del M
    
    # ==================================================================
    # Calculate gradients : cylindrical [r,θ]
    # ==================================================================
    
    ## [x1,x2] = [r,θ]
    ## ∂(A)/∂x1 = ∂(A)/∂(q1)·∂(q1)/∂(x1) + ∂(A)/∂(q2)·∂(q2)/∂(x1)
    ## ∂(A)/∂x2 = ∂(A)/∂(q1)·∂(q1)/∂(x2) + ∂(A)/∂(q2)·∂(q2)/∂(x2)
    
    A = ur
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddr_ur = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddt_ur = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    A = ut
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddr_ut = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddt_ut = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    A = p
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddr_p = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddt_p = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    # ==================================================================
    # Construct momentum balance : cylindrical [r,θ]
    # ==================================================================
    
    ## τ_rr, τ_θθ, τ_rθ
    tau_rr = +(4/3)*mu*ddr_ur - (2/3)*(1/r)*mu*ddt_ut - (2/3)*mu*(ur/r)
    tau_tt = -(2/3)*mu*ddr_ur + (4/3)*(1/r)*mu*ddt_ut + (4/3)*mu*(ur/r)
    tau_rt = mu*( (1/r)*ddt_ur + ddr_ut - (ut/r) )
    
    # === ∂r[ρ·ur·ur]
    
    A = rho * ur_Fv * ur_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t1r = ddx1_A
    
    # === (1/r)·∂θ[ρ·ur·uθ]
    
    A = rho * ur_Fv * ut_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t2r = (1/r)*ddx2_A
    
    # === ∂r[ρ·ur·uθ]
    
    A = rho * ur_Fv * ut_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t1t = ddx1_A
    
    # === (1/r)·∂θ[ρ·uθ·uθ]
    
    A = rho * ut_Fv * ut_Fv
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t2t = (1/r)*ddx2_A
    
    # === ∂r[p] & (1/r)·∂θ[p]
    
    t3r = ddr_p
    t3t = (1/r)*ddt_p
    
    # === -∂r[τ_rr]
    
    A = tau_rr
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t4r = -1 * ddx1_A
    
    # === -(1/r)·∂θ[τ_θθ]
    
    A = tau_tt
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t6t = -(1/r)*ddx2_A
    
    # === ∂r[ρur″ur″]
    
    A = r_urII_urII
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    #ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t5r = ddx1_A
    
    # === -(1/r)·∂θ[τ_rθ]  &  -∂r[τ_rθ]
    
    A = tau_rt
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t4t = -ddx1_A
    t6r = -(1/r)*ddx2_A
    
    # === (1/r)·∂θ[ρur″uθ″] & ∂r[ρur″uθ″]
    
    A = r_urII_utII
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t7r = (1/r)*ddx2_A
    t5t = ddx1_A
    
    # === (1/r)·∂θ[ρuθ″uθ″]
    
    A = r_utII_utII
    ddq1_A = gradient(A, 1., axis=0, acc=acc, edge_stencil=edge_stencil, d=1)
    ddq2_A = gradient(A, 1., axis=1, acc=acc, edge_stencil=edge_stencil, d=1)
    #ddx1_A = ddq1_A*ddx1_q1 + ddq2_A*ddx1_q2
    ddx2_A = ddq1_A*ddx2_q1 + ddq2_A*ddx2_q2
    
    t7t = (1/r)*ddx2_A
    
    # ===
    
    ## +(1/r)·ρ·ur·ur
    t8r = (1/r) * rho * ur_Fv * ur_Fv
    
    ## +(1/r)·ρur″ur″
    t9r = (1/r) * r_urII_urII
    
    ## -(1/r)·ρ·uθ·uθ
    t10r = -1 * (1/r) * rho * ut_Fv * ut_Fv
    
    ## -(1/r)·ρuθ″uθ″
    t11r = -1 * (1/r) * r_utII_utII
    
    ## -(1/r)·τ_rr
    t12r = -1 * (1/r) * tau_rr
    
    ## +(1/r)·τ_θθ
    t13r = (1/r) * tau_tt
    
    # ===
    
    ## +(2/r)·ρ·ur·uθ
    t8t = (2/r) * rho * ur_Fv * ut_Fv
    
    ## +(2/r)·ρur″uθ″
    t9t = (2/r) * r_urII_utII
    
    ## -(2/r)·τ_rθ
    t10t = -(2/r) * tau_rt
    
    # ===
    
    data_r = {
        't1':t1r,
        't2':t2r,
        't3':t3r,
        't4':t4r,
        't5':t5r,
        't6':t6r,
        't7':t7r,
        't8':t8r,
        't9':t9r,
        't10':t10r,
        't11':t11r,
        't12':t12r,
        't13':t13r,
        }
    
    data_t = {
        't1':t1t,
        't2':t2t,
        't3':t3t,
        't4':t4t,
        't5':t5t,
        't6':t6t,
        't7':t7t,
        't8':t8t,
        't9':t9t,
        't10':t10t,
        }
    
    return data_r, data_t

def _calc_momentum_balance_terms(self, **kwargs):
    '''
    Calculate terms of the 2D time-averaged momentum conservation equation
    '''
    
    verbose      = kwargs.get('verbose',True)
    local_csys   = kwargs.get('local_csys',True)
    acc          = kwargs.get('acc',6)
    edge_stencil = kwargs.get('edge_stencil','full')
    xis          = kwargs.get('xis',None) ## which [x] indices to compute (None=all)
    csys_type    = kwargs.get('csys_type','cartesian')
    
    if verbose: print('\n'+'ztmd.calc_momentum_balance_terms()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if self.open_mode not in ['a','w','r+']:
        raise ValueError('Not able to write to hdf5 file')
    
    if verbose: even_print('local_csys',str(local_csys))
    if verbose: even_print('csys_type',csys_type)
    
    if local_csys:
        if self.rectilinear:
            print('>>> WARNING: this grid is rectilinear. Using local csys is not needed.')
        if 'csys/vtang' not in self:
            raise ValueError('csys/vtang not in hdf5')
        if 'csys/vnorm' not in self:
            raise ValueError('csys/vnorm not in hdf5')
        
        ## Read unit vectors (wall tangent, wall norm) from HDF5
        vtang = np.copy( self['csys/vtang'][()] )
        vnorm = np.copy( self['csys/vnorm'][()] )
        if verbose: even_print('vtang',str(vtang.shape))
        if verbose: even_print('vnorm',str(vnorm.shape))
    
    if (csys_type=='cartesian'):
        gnx = 'data_momentum_x'
        gny = 'data_momentum_y'
        gns = [gnx,gny]
    elif (csys_type=='cylindrical'):
        gnr = 'data_momentum_r'
        gnt = 'data_momentum_theta'
        gns = [gnr,gnt]
    else:
        raise NotImplementedError
    
    terms_x = {
        't1':( r'∂x[ρ·u·u]' , r'$\partial_{x}[\bar{\rho} \tilde{u} \tilde{u}]$'),
        't2':( r'∂y[ρ·u·v]' , r'$\partial_{y}[\bar{\rho} \tilde{u} \tilde{v}]$'),
        't3':( r'∂x[p]'     , r'$\partial_{x}[\bar{p}]$'),
        't4':( r'-∂x[τxx]'  , r'$-\partial_{x}[\overline{\tau}_{xx}]$'),
        't5':( r'∂x[ρu″u″]' , r'$\partial_{x}[\overline{\rho u^{\prime\prime} u^{\prime\prime}}]$'),
        't6':( r'-∂y[τxy]'  , r'$-\partial_{y}[\overline{\tau}_{xy}]$'),
        't7':( r'∂y[ρu″v″]' , r'$\partial_{y}[\overline{\rho u^{\prime\prime} v^{\prime\prime}}]$'),
        }
    
    terms_y = {
        't1':( r'∂x[ρ·u·v]' , r'$\partial_{x}[\bar{\rho} \tilde{u} \tilde{v}]$'),
        't2':( r'∂y[ρ·v·v]' , r'$\partial_{y}[\bar{\rho} \tilde{v} \tilde{v}]$'),
        't3':( r'∂y[p]'     , r'$\partial_{y}[\bar{p}]$'),
        't4':( r'-∂x[τxy]'  , r'$-\partial_{x}[\overline{\tau}_{xy}]$'),
        't5':( r'∂x[ρu″v″]' , r'$\partial_{x}[\overline{\rho u^{\prime\prime} v^{\prime\prime}}]$'),
        't6':( r'-∂y[τyy]'  , r'$-\partial_{y}[\overline{\tau}_{yy}]$'),
        't7':( r'∂y[ρv″v″]' , r'$\partial_{y}[\overline{\rho v^{\prime\prime} v^{\prime\prime}}]$'),
        }
    
    terms_r = {
        't1':( r'∂r[ρ·ur·ur]'       , r'$\partial_{r}[ \bar{\rho} \tilde{u}_r \tilde{u}_r ]$'),
        't2':( r'(1/r)·∂θ[ρ·ur·uθ]' , r'$\frac{1}{r} \partial_{\theta}[ \bar{\rho} \tilde{u}_r \tilde{u}_\theta ]$'),
        't3':( r'∂r[p]'             , r'$\partial_{r}[\bar{p}]$'),
        't4':( r'-∂r[τ_rr]'         , r'$-\partial_{r}[ \overline{\tau}_{r r} ]$'),
        't5':( r'∂r[ρur″ur″]'       , r'$\partial_{r}[ \overline{ \rho u^{\prime \prime}_r  u^{\prime \prime}_r } ]$'),
        't6':( r'-(1/r)·∂θ[τ_rθ]'   , r'$-\frac{1}{r} \partial_{\theta}[ \overline{\tau}_{r \theta} ]$'),
        't7':( r'(1/r)·∂θ[ρur″uθ″]' , r'$\frac{1}{r} \partial_{\theta}[ \overline{\rho u^{\prime \prime}_r  u^{\prime \prime}_\theta } ]$'),
        't8':( r'(1/r)·ρ·ur·ur'     , r'$\frac{1}{r} \bar{\rho} \tilde{u}_{r} \tilde{u}_{r}$'),
        't9':( r'(1/r)·ρur″ur″'     , r'$\frac{1}{r} \overline{ \rho u^{\prime \prime}_r u^{\prime \prime}_r }$'),
        't10':( r'-(1/r)·ρ·uθ·uθ'   , r'$-\frac{1}{r} \bar{\rho} \tilde{u}_{\theta} \tilde{u}_{\theta}$'),
        't11':( r'-(1/r)·ρuθ″uθ″'   , r'$-\frac{1}{r} \overline{ \rho u^{\prime \prime}_\theta u^{\prime \prime}_\theta }$'),
        't12':( r'-(1/r)·τ_rr'      , r'$-\frac{1}{r} \overline{\tau}_{r r}$'),
        't13':( r'(1/r)·τ_θθ'       , r'$\frac{1}{r} \overline{\tau}_{\theta \theta}$'),
        }
    
    terms_theta = {
        't1':( r'∂r[ρ·ur·uθ]'       , r'$\partial_{r}[ \bar{\rho} \tilde{u}_r \tilde{u}_\theta ]$'),
        't2':( r'(1/r)·∂θ[ρ·uθ·uθ]' , r'$\frac{1}{r} \partial_{\theta}[ \bar{\rho} \tilde{u}_\theta \tilde{u}_\theta ]$'),
        't3':( r'(1/r)·∂θ[p]'       , r'$\frac{1}{r} \partial_{\theta}[\bar{p}]$'),
        't4':( r'-∂r[τ_rθ]'         , r'$-\partial_{r}[ \overline{\tau}_{r \theta} ]$'),
        't5':( r'∂r[ρur″uθ″]'       , r'$\partial_{r}[ \overline{\rho u^{\prime \prime}_r  u^{\prime \prime}_\theta } ]$'),
        't6':( r'-(1/r)·∂θ[τ_θθ]'   , r'$-\frac{1}{r} \partial_{\theta}[ \overline{\tau}_{\theta \theta} ]$'),
        't7':( r'(1/r)·∂θ[ρuθ″uθ″]' , r'$\frac{1}{r} \partial_{\theta}[ \overline{\rho u^{\prime \prime}_\theta  u^{\prime \prime}_\theta } ]$'),
        't8':( r'(2/r)·ρ·ur·uθ'     , r'$\frac{2}{r} \bar{\rho} \tilde{u}_{r} \tilde{u}_{\theta}$'),
        't9':( r'(2/r)·ρur″uθ″'     , r'$\frac{2}{r} \overline{ \rho u^{\prime \prime}_r u^{\prime \prime}_\theta }$'),
        't10':( r'-(2/r)·τ_rθ'      , r'$-\frac{2}{r} \overline{\tau}_{r \theta}$'),
        }
    
    ## Create groups in ZTMD
    for gn in gns:
        if verbose: print(72*'-')
        if gn in self:
            del self[gn]
            if verbose: even_print('deleted group',f'/{gn}/')
        
        grp = self.create_group(gn)
        if verbose: even_print('created group',f'{grp.name}/')
        
        ## Set group attributes
        grp.attrs['local_csys'] = local_csys
        if verbose: even_print(f'{gn}.local_csys',str(grp.attrs['local_csys']))
    
    gn=None; del gn
    
    if verbose: print(72*'-')
    
    ## Initialize dsets for [x] terms
    if csys_type=='cartesian':
        for dsn, tags in terms_x.items():
            unicode,latex = tags
            shape = (self.ny,self.nx)
            if (f'{gnx}/{dsn}' in self):
                del self[f'{gnx}/{dsn}']
            dset = self.create_dataset(
                                    f'{gnx}/{dsn}', 
                                    shape=shape, 
                                    dtype=np.float32,
                                    chunks=None,
                                    fillvalue=np.float32(np.nan),
                                    )
            if verbose: even_print(f'{gnx}/{dsn}',str(shape))
            dset.attrs['latex'] = latex
            dset.attrs['unicode'] = unicode
            #if verbose: even_print(f'{dsn}.latex',str(dset.attrs['latex']))
        if verbose: print(72*'-')
    
    ## Initialize dsets for [y] terms
    if csys_type=='cartesian':
        for dsn, tags in terms_y.items():
            unicode,latex = tags
            shape = (self.ny,self.nx)
            if (f'{gny}/{dsn}' in self):
                del self[f'{gny}/{dsn}']
            dset = self.create_dataset(
                                    f'{gny}/{dsn}', 
                                    shape=shape, 
                                    dtype=np.float32,
                                    chunks=None,
                                    fillvalue=np.float32(np.nan),
                                    )
            if verbose: even_print(f'{gny}/{dsn}',str(shape))
            dset.attrs['latex'] = latex
            dset.attrs['unicode'] = unicode
            #if verbose: even_print(f'{dsn}.latex',str(dset.attrs['latex']))
        if verbose: print(72*'-')
    
    ## Initialize dsets for [r] terms
    if csys_type=='cylindrical':
        for dsn, tags in terms_r.items():
            unicode,latex = tags
            shape = (self.ny,self.nx)
            if (f'{gnr}/{dsn}' in self):
                del self[f'{gnr}/{dsn}']
            dset = self.create_dataset(
                                    f'{gnr}/{dsn}', 
                                    shape=shape, 
                                    dtype=np.float32,
                                    chunks=None,
                                    fillvalue=np.float32(np.nan),
                                    )
            if verbose: even_print(f'{gnr}/{dsn}',str(shape))
            dset.attrs['latex'] = latex
            dset.attrs['unicode'] = unicode
            #if verbose: even_print(f'{dsn}.latex',str(dset.attrs['latex']))
        if verbose: print(72*'-')
    
    ## Initialize dsets for [θ] terms
    if csys_type=='cylindrical':
        for dsn, tags in terms_theta.items():
            unicode,latex = tags
            shape = (self.ny,self.nx)
            if (f'{gnt}/{dsn}' in self):
                del self[f'{gnt}/{dsn}']
            dset = self.create_dataset(
                                    f'{gnt}/{dsn}', 
                                    shape=shape, 
                                    dtype=np.float32,
                                    chunks=None,
                                    fillvalue=np.float32(np.nan),
                                    )
            if verbose: even_print(f'{gnt}/{dsn}',str(shape))
            dset.attrs['latex'] = latex
            dset.attrs['unicode'] = unicode
            #if verbose: even_print(f'{dsn}.latex',str(dset.attrs['latex']))
        if verbose: print(72*'-')
    
    ## Read data
    if csys_type=='cylindrical' and hasattr(self,'crv_R'):
        crv_R = np.copy(self.crv_R)
        if verbose:
            even_print('crv_R',str(crv_R.shape))
            print(72*'-')
    
    if verbose:
        print('Reading data'+'\n'+'------------')
    u   = np.copy( self['data/u'][()].T   )
    v   = np.copy( self['data/v'][()].T   )
    rho = np.copy( self['data/rho'][()].T )
    p   = np.copy( self['data/p'][()].T   )
    mu  = np.copy( self['data/mu'][()].T  )
    
    if verbose: even_print('u',str(u.shape))
    if verbose: even_print('v',str(v.shape))
    if verbose: even_print('ρ',str(rho.shape))
    if verbose: even_print('p',str(p.shape))
    if verbose: even_print('μ',str(mu.shape))
    
    u_Fv = np.copy( self['data/u_Fv'][()].T )
    v_Fv = np.copy( self['data/v_Fv'][()].T )
    if verbose: even_print('u_Fv',str(u_Fv.shape))
    if verbose: even_print('v_Fv',str(v_Fv.shape))
    
    r_uII_uII = np.copy( self['data/r_uII_uII'][()].T )
    r_vII_vII = np.copy( self['data/r_vII_vII'][()].T )
    r_uII_vII = np.copy( self['data/r_uII_vII'][()].T )
    if verbose: even_print('ρu″u″',str(r_uII_uII.shape))
    if verbose: even_print('ρu″v″',str(r_uII_vII.shape))
    if verbose: even_print('ρv″v″',str(r_vII_vII.shape))
    
    x = np.copy(self.x)
    y = np.copy(self.y)
    
    if x.ndim==1 and y.ndim==1:
        x,y = np.meshgrid(x,y, indexing='ij')
    if verbose: even_print('x',str(x.shape))
    if verbose: even_print('y',str(y.shape))
    if verbose: print(72*'-')
    
    nx,ny = x.shape
    
    # ==================================================================
    
    def _get_transformation_matrix(xi):
        '''
        Get transformation matrix : translate,rotate,[translate]
        '''
        rotation_angle_z = np.arcsin(vnorm[xi,0,0])
        
        rotation_matrix = np.array( [[ np.cos(rotation_angle_z) , -np.sin(rotation_angle_z) , 0 ],
                                     [ np.sin(rotation_angle_z) ,  np.cos(rotation_angle_z) , 0 ],
                                     [ 0                        ,  0                        , 1 ]], dtype=np.float64 )
        
        ## Translate to [cx,cy]
        cx = x[xi,0]
        cy = y[xi,0]
        translation_matrix_A = np.array( [[ 1 , 0 , -cx ],
                                          [ 0 , 1 , -cy ],
                                          [ 0 , 0 ,  1  ]], dtype=np.float64 )
        
        ## For cylindrical csys, translate in [y] by -R (the local curvature radius)
        ##   such that origin is distance 'R' from the wall
        if (csys_type=='cylindrical'):
            cx = 0.
            try:
                cy = crv_R[xi]
            except NameError:
                cy = 0.
            
            translation_matrix_B = np.array( [[ 1 , 0 ,  0  ],
                                              [ 0 , 1 , -cy ],
                                              [ 0 , 0 ,  1  ]], dtype=np.float64 )
        else:
            translation_matrix_B = np.identity(3, dtype=np.float64) ## No second translation
        
        ## Form the full [3x3] matrix by multiplying together
        ##  --> ORDER MATTERS!
        transformation_matrix = np.einsum( 'ij,jk,kl->il', translation_matrix_B , rotation_matrix, translation_matrix_A )
        
        return transformation_matrix
    
    # ==================================================================
    
    if local_csys: ## Compute momentum balance in local csys
        
        if xis is None: ## Do all
            xis = np.arange(self.nx)
            n_prog = self.nx
        else: ## Do only for selected x-indices
            n_prog = len(xis)
        
        ## Assert passed 'xis' is OK
        if isinstance(xis,np.ndarray): ## numpy array
            if not xis.ndim==1:
                raise ValueError('xis must have .ndim==1')
            if not np.issubdtype(xis.dtype,np.integer):
                raise RuntimeError("xis must have subtype np.integer")
        else: ## some other iterable... are all elements ints?
            try:
                xis = [int(xi) for xi in xis]
            except Exception as e:
                raise RuntimeError("elements of xis are not interpretable as ints") from e
        
        # ==============================================================
        # Main loop
        # ==============================================================
        
        if csys_type=='cartesian':
            desc='Momentum balance [x,y]'
        elif csys_type=='cylindrical':
            desc='Momentum balance [r,θ]'
        else:
            raise ValueError
        
        if verbose:
            progress_bar = tqdm(
                total=n_prog,
                ncols=100,
                desc=desc,
                leave=True,
                file=sys.stdout,
                mininterval=0.1,
                smoothing=0.,
                #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
                bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
                ascii="░█",
                colour='#00ffff',
                )
        
        for xi in xis:
            
            ## Get transformation matrix (global-->local) for this [x] index
            transformation_matrix = _get_transformation_matrix(xi)
            
            ## Take N points in [x] on either side
            ## This allows for full FD stencil at central point
            xiA = max( 0  , xi-12   )
            xiB = min( nx , xi+12+1 )
            xii = xi - xiA ## Index local to slice
            #tqdm.write(f'xi={xi:d},xiA={xiA:d},xiB={xiB:d},xii={xii:d}') ## DEBUG
            
            if (csys_type=='cartesian'):
                
                data_x, data_y = momentum_balance_kernel_cartesian(
                                                        x=x[xiA:xiB,:],
                                                        y=y[xiA:xiB,:],
                                                        u=u[xiA:xiB,:],
                                                        v=v[xiA:xiB,:],
                                                        u_Fv=u_Fv[xiA:xiB,:],
                                                        v_Fv=v_Fv[xiA:xiB,:],
                                                        p=p[xiA:xiB,:],
                                                        rho=rho[xiA:xiB,:],
                                                        mu=mu[xiA:xiB,:],
                                                        r_uII_uII=r_uII_uII[xiA:xiB,:],
                                                        r_uII_vII=r_uII_vII[xiA:xiB,:],
                                                        r_vII_vII=r_vII_vII[xiA:xiB,:],
                                                        transformation_matrix=transformation_matrix,
                                                        acc=acc,
                                                        edge_stencil=edge_stencil,
                                                        )
                
                ## Write to HDF5
                for dsn, arr in data_x.items():
                    dset = self[f'{gnx}/{dsn}']
                    dset[:,xi] = arr[xii,:] ## Write only 1 [x] position
                for dsn, arr in data_y.items():
                    dset = self[f'{gny}/{dsn}']
                    dset[:,xi] = arr[xii,:] ## Write only 1 [x] position
            
            elif (csys_type=='cylindrical'):
                
                data_r, data_t = momentum_balance_kernel_cylindrical(
                                                        x=x[xiA:xiB,:],
                                                        y=y[xiA:xiB,:],
                                                        u=u[xiA:xiB,:],
                                                        v=v[xiA:xiB,:],
                                                        u_Fv=u_Fv[xiA:xiB,:],
                                                        v_Fv=v_Fv[xiA:xiB,:],
                                                        p=p[xiA:xiB,:],
                                                        rho=rho[xiA:xiB,:],
                                                        mu=mu[xiA:xiB,:],
                                                        r_uII_uII=r_uII_uII[xiA:xiB,:],
                                                        r_uII_vII=r_uII_vII[xiA:xiB,:],
                                                        r_vII_vII=r_vII_vII[xiA:xiB,:],
                                                        transformation_matrix=transformation_matrix,
                                                        acc=acc,
                                                        edge_stencil=edge_stencil,
                                                        )
                
                ## Write to HDF5
                for dsn, arr in data_r.items():
                    dset = self[f'{gnr}/{dsn}']
                    dset[:,xi] = arr[xii,:] ## Write only 1 [x] position
                for dsn, arr in data_t.items():
                    dset = self[f'{gnt}/{dsn}']
                    dset[:,xi] = arr[xii,:] ## Write only 1 [x] position
            
            else:
                raise ValueError
            
            if verbose: progress_bar.update()
        if verbose: progress_bar.close()
    
    else: ## Compute momentum balance in global csys
        
        ## Send to kernel
        if csys_type=='cartesian': ## Cartesian
            
            data_x, data_y = momentum_balance_kernel_cartesian(
                                                        x=x,
                                                        y=y,
                                                        u=u,
                                                        v=v,
                                                        u_Fv=u_Fv,
                                                        v_Fv=v_Fv,
                                                        p=p,
                                                        rho=rho,
                                                        mu=mu,
                                                        r_uII_uII=r_uII_uII,
                                                        r_uII_vII=r_uII_vII,
                                                        r_vII_vII=r_vII_vII,
                                                        #transformation_matrix=transformation_matrix,
                                                        acc=acc,
                                                        edge_stencil=edge_stencil,
                                                        )
            
            ## Write to HDF5
            for dsn,arr in data_x.items():
                dset = self[f'{gnx}/{dsn}']
                if verbose:
                    even_print(dset.name,str(arr.T.shape))
                dset[:,:] = arr.T
            for dsn,arr in data_y.items():
                dset = self[f'{gny}/{dsn}']
                if verbose:
                    even_print(dset.name,str(arr.T.shape))
                dset[:,:] = arr.T
        
        elif csys_type=='cylindrical': ## Cylindrical
            
            data_r, data_t = momentum_balance_kernel_cylindrical(
                                                            x=x,
                                                            y=y,
                                                            u=u,
                                                            v=v,
                                                            u_Fv=u_Fv,
                                                            v_Fv=v_Fv,
                                                            p=p,
                                                            rho=rho,
                                                            mu=mu,
                                                            r_uII_uII=r_uII_uII,
                                                            r_uII_vII=r_uII_vII,
                                                            r_vII_vII=r_vII_vII,
                                                            #transformation_matrix=transformation_matrix,
                                                            acc=acc,
                                                            edge_stencil=edge_stencil,
                                                            )
            
            ## Write to HDF5
            for dsn,arr in data_r.items():
                dset = self[f'{gnr}/{dsn}']
                if verbose:
                    even_print(dset.name,str(arr.T.shape))
                dset[:,:] = arr.T
            for dsn,arr in data_t.items():
                dset = self[f'{gnt}/{dsn}']
                if verbose:
                    even_print(dset.name,str(arr.T.shape))
                dset[:,:] = arr.T
        
        else:
            raise ValueError
    
    self.get_header(verbose=False)
    if verbose: print(72*'-')
    if verbose: print('total time : ztmd.calc_momentum_balance_terms() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return
